@@ -1,5 +1,5 @@
 /*
- *  RODDC/src/RODBC.c by M. Lapsley and B. D. Ripley  Copyright (C) 1999-2011
+ *  RODDC/src/RODBC.c by M. Lapsley and B. D. Ripley  Copyright (C) 1999-2013
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -277,7 +277,7 @@ SEXP RODBCDriverConnect(SEXP connection, SEXP id, SEXP useNRows, SEXP ReadOnly)
 	    return ans;
 	} else {
 	    if (retval == SQL_ERROR) {
-		SQLCHAR state[5], msg[1000];
+		SQLCHAR state[6], msg[1000];
 		SQLSMALLINT buffsize=1000, msglen, i=1;
 		SQLINTEGER code;
 		while(1) {
@@ -1273,7 +1273,7 @@ geterr(pRODBCHandle thisHandle)
 	if(retval != SQL_SUCCESS && retval != SQL_SUCCESS_WITH_INFO) break;
 	sprintf(message,"%s %d %s", sqlstate, (int)NativeError, msg);
 	errlistAppend(thisHandle, message);
-	i++;
+	// i++;
     }
 }
 
@@ -1527,9 +1527,9 @@ static const R_CallMethodDef CallEntries[] = {
     {NULL, NULL, 0}
 };
 
-
 void R_init_RODBC(DllInfo *dll)
 {
     R_registerRoutines(dll, NULL, CallEntries, NULL, NULL);
     R_useDynamicSymbols(dll, FALSE);
+    R_forceSymbols(dll, TRUE);
 }
